package ap.proof.tree;

import ap.proof.goal.{PrioritisedTask, Goal, TaskManager, CompoundFormulas}
import ap.terfor.{Formula, ConstantTerm, TermOrder}
import ap.terfor.conjunctions.{Conjunction, NegatedConjunctions, Quantifier}
import ap.terfor.substitutions.Substitution

abstract class ProofTreeFactory {

  def and(subtrees : Seq[ProofTree], vocabulary : Vocabulary) : ProofTree
  
  def quantify(subtree : ProofTree,
               quan : Quantifier,
               quantifiedConstants : Set[ConstantTerm],
               vocabulary : Vocabulary,
               subtreeOrder : TermOrder) : ProofTree
  
  def weaken(subtree : ProofTree,
             disjunct : Conjunction,
             vocabulary : Vocabulary) : ProofTree

  /**
   * A callback to tell that a constant has been eliminated. Upon elimination,
   * it must be possible to provide a witness, i.e., given a substitution that
   * describes concrete values for all the remaining variables, a solution
   * for <code>c</code> must be computable.
   */
  def eliminatedConstant(subtree : ProofTree,
                         c : ConstantTerm,
                         witness : (Substitution, TermOrder) => Substitution,
                         vocabulary : Vocabulary) : ProofTree
             
  //////////////////////////////////////////////////////////////////////////////
  
  def updateGoal(updatedFacts : Conjunction,
                 updatedCompoundFormulas : CompoundFormulas,
                 updatedElimConstants : Set[ConstantTerm],
                 updatedVocabulary : Vocabulary,
                 updatedDefinedSyms : Substitution,
                 newTasks : Iterable[PrioritisedTask],
                 goal : Goal) : ProofTree

  def updateGoal(updatedElimConstants : Set[ConstantTerm],
                 updatedVocabulary : Vocabulary,
                 newTasks : Iterable[PrioritisedTask],
                 goal : Goal) : ProofTree =
    updateGoal(goal.facts, goal.compoundFormulas, updatedElimConstants,
               updatedVocabulary, goal.definedSyms sortBy updatedVocabulary.order,
               newTasks, goal)

  def updateGoal(updatedFacts : Conjunction,
                 newTasks : Iterable[PrioritisedTask],
                 goal : Goal) : ProofTree =
    updateGoal(updatedFacts, goal.compoundFormulas, goal.eliminatedConstants,
               goal.vocabulary, goal.definedSyms, newTasks, goal)

  def updateGoal(updatedFacts : Conjunction, goal : Goal) : ProofTree =
    updateGoal(updatedFacts, goal.compoundFormulas, goal.eliminatedConstants,
               goal.vocabulary, goal.definedSyms, List(), goal)

  def updateGoalAddQFClause(clause : Conjunction, goal : Goal) : ProofTree = {
    val newClauses = NegatedConjunctions(goal.compoundFormulas.qfClauses.elements ++
                                         Iterator.single(clause),
                                         goal.order)
    updateGoal(goal.compoundFormulas updateQFClauses newClauses, goal)
  }
  
  def updateGoal(updatedCompoundFormulas : CompoundFormulas,
                 goal : Goal) : ProofTree =
    updateGoal(goal.facts, updatedCompoundFormulas, goal.eliminatedConstants,
               goal.vocabulary, goal.definedSyms, List(), goal)

  def updateGoal(updatedCompoundFormulas : CompoundFormulas,
                 newTasks : Iterable[PrioritisedTask],
                 goal : Goal) : ProofTree =
    updateGoal(goal.facts, updatedCompoundFormulas, goal.eliminatedConstants,
               goal.vocabulary, goal.definedSyms, newTasks, goal)

  def updateGoal(updatedCompoundFormulas : CompoundFormulas,
                 newTasks : Iterable[PrioritisedTask],
                 updatedVocabulary : Vocabulary,
                 goal : Goal) : ProofTree =
    updateGoal(goal.facts, updatedCompoundFormulas, goal.eliminatedConstants,
               updatedVocabulary, goal.definedSyms, newTasks, goal)

  def updateGoal(newTasks : Iterable[PrioritisedTask], goal : Goal) : ProofTree =
    updateGoal(goal.facts, goal.compoundFormulas, goal.eliminatedConstants,
               goal.vocabulary, goal.definedSyms, newTasks, goal)

  def updateGoal(goal : Goal) : ProofTree =
    updateGoal(goal.facts, goal.compoundFormulas, goal.eliminatedConstants,
               goal.vocabulary, goal.definedSyms, List(), goal)

  /**
   * Possibly exchange the whole task manager or update arbitrary tasks
   */
  def updateGoal(tasks : TaskManager, goal : Goal) : ProofTree

}
