package ap.proof.tree;

import ap.terfor.{TermOrder, ConstantTerm, VariableTerm, Formula}
import ap.terfor.conjunctions.{Conjunction, Quantifier}
import ap.util.{PlainRange, Debug, Seqs}

object QuantifiedTree {
  
  private val AC = Debug.AC_PROOF_TREE
  
  def apply(subtree : ProofTree,
            quan : Quantifier,
            quantifiedConstants : Set[ConstantTerm],
            vocabulary : Vocabulary,
            subtreeOrder : TermOrder,
            simplifier : ConstraintSimplifier) : ProofTree = subtree match {
    case subtree : QuantifiedTree if (subtree.quan == quan) =>
      QuantifiedTree(subtree.subtree,
                     quan,
                     quantifiedConstants ++ subtree.quantifiedConstants,
                     vocabulary, subtree.subtreeOrder,
                     simplifier)
    case subtree : AndTree if (quan == Quantifier.ALL) =>
      AndTree(for (t <- subtree.subtrees)
              yield QuantifiedTree(t, quan, quantifiedConstants, vocabulary,
                                   subtreeOrder, simplifier),
              vocabulary, simplifier)
    case subtree : WeakenTree if (quan == Quantifier.EX) =>
      WeakenTree(QuantifiedTree(subtree.subtree,
                                quan, quantifiedConstants,
                                vocabulary, subtreeOrder, simplifier),
                 quantifySimplify(subtree.disjunct, quan, quantifiedConstants,
                                  vocabulary.order, subtreeOrder, simplifier),
                 vocabulary)
    case subtree : WeakenTree if (Seqs.disjoint(quantifiedConstants,
                                                subtree.disjunct.constants)) =>
      WeakenTree(QuantifiedTree(subtree.subtree,
                                quan, quantifiedConstants,
                                vocabulary, subtreeOrder, simplifier),
                 vocabulary.order sort subtree.disjunct,
                 vocabulary)
    case _ => 
      new QuantifiedTree(subtree, quan, quantifiedConstants, vocabulary, subtreeOrder,
                         simplifier)
  }
 
  def unapply(tree : ProofTree) : Option[(Quantifier, Set[ConstantTerm], ProofTree)] =
    tree match {
      case tree : QuantifiedTree => Some(tree.quan, tree.quantifiedConstants, tree.subtree)
      case _ => None
    }
  
  private def quantify(f : Conjunction,
                       quan : Quantifier, quantifiedConstants : Set[ConstantTerm],
                       outputOrder : TermOrder, inputOrder : TermOrder) : Conjunction =
    Conjunction.quantify(quan, quantifiedConstants,
                         f, inputOrder) sortBy outputOrder
  
  private def quantifySimplify(f : Conjunction,
                               quan : Quantifier, quantifiedConstants : Set[ConstantTerm],
                               outputOrder : TermOrder, inputOrder : TermOrder,
                               simplifier : ConstraintSimplifier) : Conjunction =  
    simplifier(quantify(f, quan, quantifiedConstants, outputOrder, inputOrder),
               outputOrder)
  
}

/**
 * <code>ProofTreeOneChild</code> that quantifies a set of constants in
 * the closing constraint of its <code>subtree</code>
 */
class QuantifiedTree private (val subtree : ProofTree,
                              val quan : Quantifier,
                              val quantifiedConstants : Set[ConstantTerm],
                              val vocabulary : Vocabulary,
                              protected val subtreeOrder : TermOrder,
                              simplifier : ConstraintSimplifier)
      extends ProofTreeOneChild {

  //////////////////////////////////////////////////////////////////////////////
  Debug.assertCtor(QuantifiedTree.AC,
                   !quantifiedConstants.isEmpty &&
                   (subtree match {
                     case subtree : QuantifiedTree => subtree.quan != this.quan
                     case subtree : AndTree => this.quan != Quantifier.ALL
                     case subtree : WeakenTree if (this.quan == Quantifier.EX) =>
                       false
                     case subtree : WeakenTree if (this.quan == Quantifier.ALL) =>
                       !Seqs.disjoint(quantifiedConstants, subtree.disjunct.constants)
                     case _ => true
                    }))
  //////////////////////////////////////////////////////////////////////////////

  lazy val closingConstraint : Conjunction =
    simplifier(QuantifiedTree.quantify(subtree.closingConstraint, quan,
                                       quantifiedConstants, order, subtreeOrder),
               order)
  
  lazy val closingConstantFreedom : ConstantFreedom =
    if (quan == Quantifier.ALL)
      subtree.closingConstantFreedom -- quantifiedConstants
    else {
/*      val res = subtree.closingConstantFreedom
      if (res < constantFreedom) {
        val u = constantFreedom.unshieldedPart(closingConstraint, bindingContext)
        val c = constantFreedom.findNonFreeness(u, bindingContext)
        if (res < c)
        println("hallo")
      }
      res */
      subtree.closingConstantFreedom
    }

  lazy val fixedConstantFreedom : Boolean =
    subtree.fixedConstantFreedom && constantFreedom == closingConstantFreedom 

  lazy val stepMeaningful : Boolean = subtree.stepMeaningful

  def update(newSubtree : ProofTree,
             newConstantFreedom : ConstantFreedom) : ProofTree =
    if (subtree == newSubtree && constantFreedom == newConstantFreedom)
      this
    else
      QuantifiedTree(newSubtree, quan, quantifiedConstants,
                     vocabulary updateConstantFreedom newConstantFreedom,
                     subtreeOrder, simplifier)

  def newConstantFreedomForSubtree(cf : ConstantFreedom) : ConstantFreedom =
    if (quan == Quantifier.ALL)
      (cf addTopStatus quantifiedConstants) meet subtree.closingConstantFreedom
    else
      cf

  //////////////////////////////////////////////////////////////////////////////

  override def toString : String =
    "^ " + closingConstraint + " (" +
    quan + " " +
    (for (c <- quantifiedConstants) yield c.toString).mkString(" ") +
    ")\n" + subtree

}
